-- me_containers.e v3.0.0 by Mike Nelson
-- Requires Euphoria 2.5    
-- Method Euphoria Standard Class Library
-- Containers 
 
-- (v1.1.0) Container class: added get_key method; 
-- get_data and get_keys getter methods are now public.   

-- (v1.5.0) Type checked containers. 

-- (v2.0.0) Improved priority containers.

                
without trace
without warning

include misc.e -- for reverse()  
include method_euphoria.e    
  

function binary_search(sequence data,object target,integer searchType)
   integer min,max,index,mid,test 
   min=1
   max=length(data)
   index=0
   while min<=max do  
      mid=floor((min+max)/2)
      test=compare(target,data[mid])
      if test=0 then
         index=mid
         test=searchType
      end if
      if test>0 then
         min=mid+1
      else
         max=mid-1
      end if
   end while
   if index=0 then return -min end if
   return index
end function     


package("ME_Std_Lib")
package("Containers")  


type positive_integer(object x)
   if integer(x) then return x>0 end if
   return FALSE
end type  
register_type("positive_integer",routine_id("positive_integer")) 


global constant Container_Error=exception("Container_Error",Exception)

global constant Empty_Container=exception("Empty_Container",Container_Error)

global constant Full_Container=exception("Full_Container",Container_Error) 

global constant Locked_Container=
exception("Locked_Container",Container_Error)

global constant Duplicate_Element=
exception("Duplicate_Element",Container_Error)
  
global constant Invalid_Index=exception("Invalid_Index",Container_Error) 

function validate_index(object index,sequence data,integer plusOne)
   integer len   
   if not integer(index) then 
      throw(Type_Check_Failure)
      return NIL
   end if
   len=length(data)  
   if len=0 and plusOne=0 then
      throw(Empty_Container)
      return NIL   
   end if
   while TRUE do
      if index>len+plusOne then exit end if
      if index>0 then return index end if
      index=len+1+index  
      if index<1 then exit end if
      if index>len+plusOne then exit end if
      return index 
   end while         
   throw(Invalid_Index)
   return NIL
end function 

  
global constant General=interface("General",NONE,NONE,NONE)
 
global constant Basic=interface("Basic",General,
{"get_size","get","add","remove","clear","find","find_last",
"find_all"},
NONE)    

global constant Extension=interface("Extension",General,NONE,NONE)

global constant Sorted=interface("Sorted",Basic,NONE,NONE)

global constant Mapped=interface("Mapped",Sorted,{"get_key"},NONE)

global constant Unique=interface("Unique",Sorted,NONE,NONE)


global constant Container=class("Container",Entity,Basic)
 
property("data",INSTANCE,PUBLIC,PRIVATE,NONE) 
property("data_type",INSTANCE,PUBLIC,PRIVATE,"anything")
property("keys",INSTANCE,PUBLIC,PRIVATE,NONE)     
property("key_type",INSTANCE,PUBLIC,PRIVATE,"anything")

function Container_new(object data_type,object key_type) 
   entity new
   validate(data_type,"type_specifier",OPTIONAL)
   validate(key_type,"type_specifier",OPTIONAL)
   if failure() then return NOTHING end if
   new=call_super(NONE)
   if not equal(data_type,NOTHING) then
      set_property(new,"data_type",data_type)   
   end if
   if implements(this(),Mapped) then
      if not equal(key_type,NOTHING) then
         set_property(new,"key_type",key_type)   
      end if   
   end if
   return new
end function
method("new",CLASS,PUBLIC,2,routine_id("Container_new"))

super_method("clone",INSTANCE)

function Container_delete()
   VOID=call_method(this(),"clear",NONE)
   return call_super(NONE)   
end function              
method("delete",INSTANCE,PUBLIC,0,routine_id("Container_delete"))

function Container_get_size()
   return length(get_property(this(),"data"))
end function
method("get_size",INSTANCE,PUBLIC,0,routine_id("Container_get_size")) 

function do_find(object target,integer fl)
   integer index,lastIndex 
   sequence data
   if implements(this_class(),Mapped) then
      return binary_search(get_property(this(),"keys"),target,fl)
   elsif implements(this_class(),Sorted) then
      return binary_search(get_property(this(),"data"),target,fl)
   elsif fl=-1 then
      return find(target,get_property(this(),"data"))   
   else
      data=reverse(get_property(this(),"data"))
      index=find(target,data)
      if index=NIL then return NIL end if
      return length(data)+1-index
   end if       
end function  

function Container_find(object target)
   return do_find(target,-1)   
end function
method("find",INSTANCE,PUBLIC,1,routine_id("Container_find")) 

function Container_find_last(object target)
   return do_find(target,1)   
end function
method("find_last",INSTANCE,PUBLIC,1,routine_id("Container_find_last"))  

function Container_find_all(object target)
   integer first,index,lastIndex
   sequence data,res
   first=do_find(target,-1)
   if first<1 then return NONE end if
   res={first} 
   if implements(this_class(),Sorted) then
      for i=first+1 to do_find(target,1) do
         res&=i   
      end for  
      return res
   end if
   data=get_property(this(),"data")
   index=first 
   lastIndex=first
   while TRUE do
      index=find(target,data[index+1..$])+lastIndex
      if index=lastIndex then return res end if 
      res&=index
      lastIndex=index
   end while
end function  
method("find_all",INSTANCE,PUBLIC,1,routine_id("Container_find_all"))

function Container_get(object index)                                 
   sequence data 
   data=get_property(this(),"data")      
   index=validate_index(index,data,FALSE) 
   if failure() then return NOTHING end if
   return data[index]
end function
method("get",INSTANCE,PUBLIC,1,routine_id("Container_get"))
 
function Container_get_key(object index)                                 
   sequence keys  
   if not implements(this_class(),Mapped) then return NOTHING end if
   keys=get_property(this(),"keys")      
   index=validate_index(index,keys,FALSE) 
   if failure() then return NOTHING end if
   return keys[index]
end function
method("get_key",INSTANCE,PUBLIC,1,routine_id("Container_get_key"))

function Container_add(object newValue,object index) 
   integer sIndex
   sequence data,data_type,keys,key_type 
   data_type=get_property(this(),"data_type")
   if not equal(data_type,"anything") then
      validate(newValue,data_type,REQUIRED)   
      if failure() then return NIL end if
   end if
   data=get_property(this(),"data")
   if not implements(this_class(),Sorted) then
      index=validate_index(index,data,TRUE)
      if success() then 
         data=data[1..index-1]&{newValue}&data[index..$]
         set_property(this(),"data",data)
      end if
      return index 
   end if
   if implements(this_class(),Mapped) then 
      key_type=get_property(this(),"key_type")
      if not equal(key_type,"anything") then
         validate(index,key_type,REQUIRED)   
         if failure() then return NIL end if   
      end if
      keys=get_property(this(),"keys")
      sIndex=binary_search(keys,index,-1)
   else
      sIndex=binary_search(data,newValue,-1)   
   end if
   if sIndex>0 then
      if implements(this_class(),Unique) then
         throw(Duplicate_Element)
         return NIL   
      end if      
      sIndex+=1
   else
      sIndex=-sIndex
   end if 
   if implements(this_class(),Mapped) then
      keys=keys[1..sIndex-1]&{index}&keys[sIndex..$]
      set_property(this(),"keys",keys)   
   end if                             
   data=data[1..sIndex-1]&{newValue}&data[sIndex..$]
   set_property(this(),"data",data)
   return sIndex
end function
method("add",INSTANCE,PUBLIC,2,routine_id("Container_add")) 

function Container_remove(object index)
   sequence data,keys
   object item 
   data=get_property(this(),"data")
   index=validate_index(index,data,FALSE)
   if failure() then return NOTHING end if
   item=data[index]                       
   data=data[1..index-1]&data[index+1..$]
   set_property(this(),"data",data)
   if implements(this_class(),Mapped) then
      keys=get_property(this(),"keys")
      keys=keys[1..index-1]&keys[index+1..$]      
      set_property(this(),"keys",keys)
   end if
   return item
end function
method("remove",INSTANCE,PUBLIC,1,routine_id("Container_remove"))  

procedure Container_clear() 
   set_property(this(),"data",NONE)
   if implements(this_class(),Mapped) then
      set_property(this(),"keys",NONE)   
   end if   
end procedure 
method("clear",INSTANCE,PUBLIC,0,{routine_id("Container_clear")})  

end_class()  
           

global constant Multi_Set=class("Multi_Set",Container,Sorted) 
end_class()  


global constant Set=class("Set",Multi_Set,Unique)
end_class()


global constant Multi_Map=class("Multi_Map",Container,Mapped)

function Multi_Map_add(object key,object data)
   return call_super({data,key})   
end function                    
method("add",INSTANCE,PUBLIC,2,routine_id("Multi_Map_add"))  

end_class()


global constant Map=class("Map",Multi_Map,Unique)
end_class()


global constant Stack=class("Stack",Container,NONE)

function Stack_add(object newValue)
   return call_super({newValue,0})   
end function  
method("add",INSTANCE,PUBLIC,1,routine_id("Stack_add"))    

function Stack_remove()
   return call_super({-1})   
end function
method("remove",INSTANCE,PUBLIC,0,routine_id("Stack_remove"))

end_class() 

                                 
global constant Queue=class("Queue",Container,NONE)

method("add",INSTANCE,PUBLIC,1,routine_id("Stack_add")) 

function Queue_remove()
   return call_super({1})   
end function  
method("remove",INSTANCE,PUBLIC,0,routine_id("Queue_remove"))

end_class()                      


global constant Priority_Stack=class("Priority_Stack",Stack,Sorted) 

function Priority_Stack_find(object target)
   sequence data  
   data=call_method(this(),"get_data",NONE) 
   for i=1 to length(data) do
      if equal(data[i][2],target) then return i end if
   end for
   return NIL   
end function 
method("find",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find"))

function Priority_Stack_find_last(object target)
   sequence data  
   data=call_method(this(),"get_data",NONE) 
   for i=length(data) to 1 by -1 do
      if equal(data[i][2],target) then return i end if
   end for
   return NIL   
end function
method("find_last",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find_last")) 

function Priority_Stack_find_all(object target)
   sequence data,res  
   data=call_method(this(),"get_data",NONE)
   res=NONE 
   for i=1 to length(data) do
      if equal(data[i][2],target) then res&={i} end if
   end for
   return res   
end function 
method("find_all",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find_all"))

function Priority_Stack_get(object index)
   object item
   item=call_super({index})
   if failure() then return NOTHING end if
   return item[2]
end function  
method("get",INSTANCE,PUBLIC,1,routine_id("Priority_Stack_get"))

function Priority_Stack_add(object newValue,object priority)
   validate(priority,"atom",OPTIONAL)
   if equal(priority,NOTHING) then priority=0 end if
   if failure() then return NOTHING end if 
   return call_super({{priority,newValue}})
end function
method("add",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_add")) 

function Priority_Stack_remove()
   object item
   item=call_super(NONE)
   if failure() then return NOTHING end if
   return item[2]
end function  
method("remove",INSTANCE,PUBLIC,0,routine_id("Priority_Stack_remove"))

end_class()


global constant Priority_Queue=class("Priority_Queue",Queue,Sorted) 

method("find",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find")) 

method("find_last",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find_last"))

method("find_all",INSTANCE,PUBLIC,2,routine_id("Priority_Stack_find_all"))

method("get",INSTANCE,PUBLIC,1,routine_id("Priority_Stack_get"))

function Priority_Queue_add(object newValue,object priority)
   validate(priority,"atom",OPTIONAL)  
   if equal(priority,NOTHING) then priority=0 end if
   if failure() then return NOTHING end if 
   return call_super({{-priority,newValue}})
end function
method("add",INSTANCE,PUBLIC,2,routine_id("Priority_Queue_add")) 

method("remove",INSTANCE,PUBLIC,0,routine_id("Priority_Stack_remove"))

end_class()   


global constant Limited=class("Limited",Entity,Extension)

property("container",INSTANCE,PRIVATE,PRIVATE,NOTHING)
property("capacity",INSTANCE,PUBLIC,PRIVATE,NOTHING)  

function Limited_new(object container,object capacity)
   entity new
   validate(container,{"instance",General},REQUIRED) 
   validate(capacity,"positive_integer",REQUIRED)
   if failure() then return NOTHING end if 
   new=call_super(NONE)
   set_property(new,"container",container)
   set_property(new,"capacity",capacity)
   return new
end function 
method("new",CLASS,PUBLIC,2,routine_id("Limited_new"))

function Limited_clone() 
   entity container,clone
   clone=call_super(NONE)
   container=get_property(clone,"container")
   container=call_method(container,"clone",NONE)
   set_property(clone,"container",container) 
   return clone
end function                                    
method("clone",INSTANCE,PUBLIC,0,routine_id("Limited_clone"))

function Limited_delete()
   entity container
   container=get_property(this(),"container")
   VOID=call_method(container,"delete",NONE)
   return call_super(NONE)   
end function             
method("delete",INSTANCE,PUBLIC,0,routine_id("Limited_delete"))

function Limited_add(sequence params)
   integer capacity
   entity container
   capacity=get_property(this(),"capacity")
   container=get_property(this(),"container")
   if call_method(container,"get_size",NONE)>=capacity then
      throw(Full_Container)
      return NOTHING   
   end if           
   return call_method(container,"add",params)
end function                                 
method("add",INSTANCE,PUBLIC,-1,routine_id("Limited_add"))   

function Limited_undefined_method(object name,object params)
   entity container
   container=get_property(this(),"container")
   return call_method(container,name,params)      
end function
method("undefined_method",INSTANCE,PROTECTED,-2,
routine_id("Limited_undefined_method"))

end_class()  


global constant Locking=class("Locking",Entity,Extension)

property("container",INSTANCE,PRIVATE,PRIVATE,NOTHING)
property("locked",INSTANCE,PUBLIC,PRIVATE,FALSE)  

function Locking_new(object container)
   entity new
   validate(container,{"instance",General},REQUIRED) 
   if failure() then return NOTHING end if 
   new=call_super(NONE)
   set_property(new,"container",container)
   return new
end function 
method("new",CLASS,PUBLIC,1,routine_id("Locking_new"))

method("clone",INSTANCE,PUBLIC,0,routine_id("Limited_clone"))

method("delete",INSTANCE,PUBLIC,0,routine_id("Limited_delete"))  

function Locking_undefined_method(object name,object params)
   entity container
   if find(name,{"add","remove","clear"}) then
      if get_property(this(),"locked") then
         throw(Locked_Container)
         if equal(name,"add") then return NIL end if
         return NOTHING                                    
      end if            
   end if
   container=get_property(this(),"container")  
   return call_method(container,name,params)
end function   
method("undefined_method",INSTANCE,PROTECTED,-2,
routine_id("Locking_undefined_method"))

procedure Locking_lock()
   set_property(this(),"locked",TRUE)   
end procedure     
method("lock",INSTANCE,PUBLIC,0,{routine_id("Locking_lock")})

procedure Locking_unlock()
   set_property(this(),"locked",FALSE)   
end procedure                         
method("unlock",INSTANCE,PUBLIC,0,{routine_id("Locking_unlock")})

end_class() 


global constant Lock_Once=class("Lock_Once",Locking,NONE)

null_method("unlock",INSTANCE,PUBLIC,NOTHING)

end_class()

end_package()
end_package()


