-- me_euphoria.e v3.0.0 by Mike Nelson  
-- Requires Euphoria 2.5  
-- Method Euphoria kernel
                              
-- A new Java-like class system for Euphoria with single inheritance and
-- interfaces.      
-- Exception handling for errors.   
-- Polymorphic methods    
-- Event handlers implemented as method wrappers (delegates). 

-- (v1.0.1) Bug fix: empty sequences serialized correctly but crashed 
-- on deserialization

-- (v1.1.0) Internal code improvement, documentation cleanup.
-- Improvements and additions to Standard Class Library containers.

-- (v1.2.0) Bug fix: type check failure in Method_Wrapper implementation
-- corrected.  Method_Wrapper's call method now takes a parameter array. 

-- (v1.3.0) Improvements to registered types, get_class() can now take a
-- a class name as a parameter.      

-- (v1.4.0) New exception model.   

-- (v1.5.0) Classes can define "undefined_method" methods to be automatically
-- called if the call of an undefined method is attempted. This facilitates
-- creating wrapper classes. New types. Standard Class Library now provides
-- type checked containers.

-- (v1.6.0) Methods may be coded as procedures.  

-- (v1.7.0) Classes may be referenced by the class name (as a string) 
-- as well as by entity handle. This can resolve namespace issues. 

-- (v2.0.0) Most run time fatal errors are now exceptions. 
-- Completely redesigned serialization.
-- Improvements and additions to the Standard Class Library. 

-- (v2.0.1) Bug fix in undefined_method() routine. Code added to prevent 
-- Method_Wrapper from wrapping an interface.

-- (v3.0.0) Implements task switching.

 
without trace
without warning  

-- CONSTANTS:
-- Entity marker:
constant MARKER=-power(2,30)
-- Classes:          
constant CLASS_FLAG=0   
constant CLASS_HEAD={MARKER,CLASS_FLAG}
-- Methods: 
global constant NOTHING=MARKER
constant BAD_METHOD=-1  
-- Meaningless return values:
global constant NIL=0
-- Booleans:
global constant FALSE=0
global constant TRUE=1  
-- Property and method varieties:
global constant INSTANCE=1
global constant CLASS=2
-- Access control:
global constant PUBLIC=3
global constant PROTECTED=4
global constant PRIVATE=5
-- Program contexts:  
constant INITIALIZATION=6
constant CLASS_DEFINITION=7  
constant METHOD=8 
-- Empty lists:
global constant NONE={}
-- Program stack:
constant STACK_EXPAND=repeat(NIL,16)
-- Identifiers:
constant ID_INIT="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
constant ID_DIGIT="0123456789"
constant ID_CHAR=ID_INIT&ID_DIGIT&'_' 
-- Validation:  
global constant REQUIRED=TRUE
global constant OPTIONAL=FALSE
constant VALID_TARGET_TYPE={INSTANCE,CLASS} 
constant VALID_ACCESS={PUBLIC,PROTECTED,PRIVATE} 
constant ENTITY_TYPES={"entity","instance","class"} 
constant SERIALIZATION_METHODS={"serialize","deserialize"} 
constant NO_DIRECT_CALLS=
{SERIALIZATION_METHODS&{"undefined_method"},
{"initialize","terminate","undefined_method"}}
constant NO_SUPER_CALLS={SERIALIZATION_METHODS,NONE}
-- Error messages:
constant CONSTANT_NAME={"instance","class","public","protected","private",
"initialization","class definition","method"}
constant CONSTANT_CAP_NAME={"Instance","Class","Public","Protected",
"Private","Initialization","Class definition","Method"} 
constant CONSTANT_UPPER_NAME={"INSTANCE","CLASS","PUBLIC","PROTECTED",
"PRIVATE","INITIALIZATION","CLASS DEFINITION","METHOD"}
-- Predefined Classes   
constant ENTITY_ID=1
global constant Entity=CLASS_HEAD&ENTITY_ID
constant INTERFACE_ID=2
global constant Interface=CLASS_HEAD&INTERFACE_ID
constant EXCEPTION_ID=3
global constant Exception=CLASS_HEAD&EXCEPTION_ID  
-- Versioning: 
global constant CURRENT_VERSION={3,0,0}  
-- Serialization:  
constant MEM=machine_func(16,4) -- allocate() 
constant VERSION_LIST={CURRENT_VERSION,{2,0,1},{2,0,0},{1,7,0},{1,6,0},{1,5,0},
{1,4,0},{1,3,0},{1,2,0},{1,1,0},{1,0,1},{1,0,0}} 
constant SV2=find({2,0,0},VERSION_LIST)  
constant OLD_NOTHING=repeat(-power(2,30),2)
-- Misc.:   
constant DOUBLE_NONE={NONE,NONE} 
constant TRIPLE_NONE=repeat(NONE,3)
constant TRIPLE_DOUBLE_NONE=repeat(DOUBLE_NONE,3)  
-- Name resolution:
constant PACKAGES={"ME_Kernel","ME_Std_Lib","Me_Std_Lib.Containers",
"ME_Std_Lib.Counted","ME_Std_Lib.File_Errors","Me_Std_Lib.Math_Errors",
"ME_Std_Lib.Misc_Errors","ME_Std_Lib.Variables","ME_Std_Lib.Widgets"}

-- VARIABLES: 
-- Trash can:
global object VOID VOID=NOTHING
-- Packages:
sequence CurrentPackage CurrentPackage=NONE
-- Classes:
integer ClassCounter ClassCounter=3 
sequence ClassName ClassName=
{"ME_Kernel.Entity","ME_Kernel.Interface","ME_Kernel.Exception"}            
sequence SuperclassList SuperclassList=
{{ENTITY_ID},{INTERFACE_ID},{EXCEPTION_ID}}
sequence ProtectedAccess ProtectedAccess={{ENTITY_ID},NONE,NONE}
sequence InterfaceList InterfaceList=TRIPLE_NONE 
sequence PropertyName PropertyName=TRIPLE_DOUBLE_NONE
sequence PropertySource PropertySource=TRIPLE_DOUBLE_NONE 
sequence PropertyType PropertyType=TRIPLE_DOUBLE_NONE
sequence PropertyDefault PropertyDefault=TRIPLE_DOUBLE_NONE
sequence MethodName MethodName=TRIPLE_DOUBLE_NONE 
sequence MethodSource MethodSource=TRIPLE_DOUBLE_NONE  
sequence MethodAccess MethodAccess=TRIPLE_DOUBLE_NONE
sequence MethodParamCount MethodParamCount=TRIPLE_DOUBLE_NONE
sequence MethodID MethodID=TRIPLE_DOUBLE_NONE 
sequence MethodReturn MethodReturn=TRIPLE_DOUBLE_NONE
-- Registered types:
sequence TypeName TypeName=NONE
sequence TypeID TypeID=NONE
-- Instances:
integer InstanceCounter InstanceCounter=1
sequence Instances Instances=repeat(NIL,128) 
sequence ClassInstances ClassInstances=TRIPLE_NONE
-- Program context:                    
integer Context Context=INITIALIZATION
integer ProgramCounter ProgramCounter=1
sequence Caller Caller=STACK_EXPAND
sequence This This=STACK_EXPAND 
sequence ThisMethod ThisMethod=STACK_EXPAND  
-- Exceptions:
sequence CurrentException CurrentException=STACK_EXPAND
-- Error handling:
sequence ErrorText ErrorText=NONE          
integer DebugFromFatal DebugFromFatal=FALSE
integer DebugOpenFail DebugOpenFail=FALSE
integer ErrorScreenWidth ErrorScreenWidth=80 
integer TerminationClass TerminationClass=NIL 
integer LastInitialize LastInitialize=NIL 
integer FatalInProgress FatalInProgress=FALSE 
integer FatalFromCrash FatalFromCrash=FALSE 
integer CrashFromFatal CrashFromFatal=FALSE
sequence FatalMessage          
-- Serialization:  
integer SVersion
integer SerInProgress SerInProgress=FALSE
integer SerErr  
integer SerFile
sequence Bytecode 
integer DesPtr
sequence InstanceList 
object ThisSer ThisSer=NOTHING 
-- Forward and special references: 
integer CALL_METHOD    
integer DEBUG_FILE 
integer DESERIALIZE_ERROR_ID
integer GET_PROPERTY  
sequence METHOD_WRAPPER 
integer METHOD_WRAPPER_ID  
integer NULL_METHOD_ID
sequence PROGRAM
integer SERIALIZE_ERROR_ID  
integer SET_PROPERTY  
integer SUPER_METHOD_ID  
integer THROW
sequence TYPE_CHECK_FAILURE  
integer VALIDATE


-- ROUTINES: 
-- Error handling:
function context_string()
   integer ce
   sequence cs
   cs=NONE
   if Context=INITIALIZATION then
      cs="In initialization." 
   elsif Context=CLASS_DEFINITION then
      cs="In "&ClassName[Caller[3]]&" class definition." 
   else 
      for i=ProgramCounter to 1 by -1 do  
         if i=ProgramCounter then cs="In " else cs&=",\ncalled from " end if    
         cs&=ClassName[Caller[i]]&" "    
         if This[i][2]!=CLASS_FLAG then
            cs&="instance"
         else
            cs&="class"  
         end if
         cs&=" method "&ThisMethod[i]
         ce=CurrentException[i]
         if ce>NIL then
            cs&=" ("&ClassName[ce]&" pending)"
         elsif ce<NIL then
            cs&=" ("&ClassName[-ce]&" caught)"   
         end if
      end for 
      if ProgramCounter>0 then cs&="." end if
   end if
   return cs
end function

type message_string(object x)
    object char
    if atom(x) then return FALSE end if
    for i=1 to length(x) do
        char=x[i]
        if not integer(char) then return FALSE end if
        if char<32 then return FALSE end if
        if char>255 then return FALSE end if
    end for
    return TRUE
end type  

type extended_message_string(object x)
   integer index
   if atom(x) then return FALSE end if
   while TRUE do
      index=find('\n',x)
      if index=NIL then exit end if
      if not message_string(x[1..index-1]) then return FALSE end if
      x=x[index+1..$]
   end while
   return message_string(x)
end type  

function wrap_line(sequence line)
   integer offset,len
   len=length(line)  
   if len=0 then return "\n" end if
   offset=0
   while len>ErrorScreenWidth-1 do 
      for i=ErrorScreenWidth-1 to 1 by -1 do
         if line[i+offset]=' ' then
            line[i+offset]='\n'
            offset+=i
            len-=i
            exit
         end if
      end for
   end while 
   if line[$]!='\n' then line&='\n' end if
   return line   
end function


global procedure error_screen_width(object width)
   if not integer(width) then return end if
   if width<20 then return end if
   ErrorScreenWidth=width
end procedure

global procedure fatal_error(object ErrorMessage)
   integer index 
   sequence errmess 
   if FatalInProgress then return end if
   if not extended_message_string(ErrorMessage) then
      ErrorMessage="*** Invalid error message. ***"
   end if                    
   ErrorText=ErrorMessage    
   errmess="METHOD EUPHORIA "&sprintf("%d.%d.%d",CURRENT_VERSION)
   &" FATAL ERROR\n"&ErrorMessage&"\n"&context_string() 
   ErrorMessage=NONE
   while length(errmess)!=0 do
      index=find('\n',errmess)
      if index=NIL then 
         ErrorMessage&=wrap_line(errmess)
         errmess=NONE
      else
         ErrorMessage&=wrap_line(errmess[1..index-1]) 
         errmess=errmess[index+1..$]
      end if
   end while    
   if DebugOpenFail=FALSE then
      DebugFromFatal=TRUE
      call_proc(DEBUG_FILE,{})
      ErrorMessage&=ErrorText&"\n"   
   end if 
   FatalMessage=ErrorMessage  
   if FatalFromCrash then
      puts(2,FatalMessage)
   else 
      CrashFromFatal=TRUE     
      machine_proc(37,FatalMessage) -- crash_message 
      ? 1/0 -- to generate ex.err    
   end if
end procedure                        

global procedure debug_file()
   integer ce,fn,source  
   sequence d,errTxt                    
   errTxt=ErrorText
   fn=open("me.err","w")
   if fn=-1 then
      DebugOpenFail=TRUE
      if DebugFromFatal then
         ErrorText="*** Cannot open me.err debug file. ***"
      end if                                               
      return
   end if
   ErrorText="New debug file me.err created."
   d=date()
   printf(fn,"--- METHOD EUPHORIA v%d.%d.%d DIAGNOSTIC LISTING FOR "
   &"%02d/%02d/%04d %02d:%02d:%02d ---",
   CURRENT_VERSION&{d[2],d[3],d[1]+1900,d[4],d[5],d[6]})
   if length(errTxt)!=0 then
      puts(fn,"\n\n\n--- FATAL ERROR ---\n\n")
      puts(fn,errTxt)
   end if
   puts(fn,"\n\n\n--- PROGRAM STACK ---\n\n") 
   if Context=INITIALIZATION then
      printf(fn,"%2d  initialization\n",{1})
   elsif Context=CLASS_DEFINITION then
      printf(fn,"%2d  %s class definition\n",{1,ClassName[Caller[3]]})   
   elsif ProgramCounter>0 then
      for i=ProgramCounter to 1 by -1 do
         printf(fn,"%2d  ",{i})
         puts(fn,ClassName[Caller[i]]&" ") 
         if This[i][2]!=CLASS_FLAG then
            puts(fn,"instance")
         else
            puts(fn,"class")   
         end if 
         puts(fn," method "&ThisMethod[i]) 
         ce=CurrentException[i]
         if ce>NIL then
            puts(fn," ("&ClassName[ce]&" pending)")
         elsif ce<NIL then
            puts(fn," ("&ClassName[-ce]&" caught)")   
         end if
         puts(fn,"\n")
      end for
   end if 
   if ClassCounter>0 then puts(fn,"\n\n--- CLASSES ---\n") end if
   for i=1 to ClassCounter do
      printf(fn,"\n\n--- %s ---\n\n",{ClassName[i]})
      printf(fn,"ID:  %d\n",i) 
      for j=2 to length(SuperclassList[i]) do
         if j=2 then
            puts(fn,"SUPERCLASSES:  ")
         else
            puts(fn,", ")
         end if
         puts(fn,ClassName[SuperclassList[i][j]])
      end for
      if length(SuperclassList[i])>1 then puts(fn,"\n") end if
      for j=1 to length(InterfaceList[i]) do
         if j=1 then
            puts(fn,"INTERFACES:  ")
         else
            puts(fn,", ")
         end if
         puts(fn,ClassName[InterfaceList[i][j]])
      end for
      if length(InterfaceList[i])>0 then puts(fn,"\n") end if  
      for j=INSTANCE to CLASS do
         for k=1 to length(PropertyName[i][j]) do 
            printf(fn,"%s PROPERTY:  %s",
            {CONSTANT_UPPER_NAME[j],PropertyName[i][j][k]})
            source=PropertySource[i][j][k]
            if source!=i then
               printf(fn,"  (%s)",{ClassName[source]})   
            end if 
            puts(fn,"\n")
         end for
      end for           
      for j=INSTANCE to CLASS do
         for k=1 to length(MethodName[i][j]) do 
            printf(fn,"%s METHOD:  %s  %s",
            {CONSTANT_UPPER_NAME[j],MethodName[i][j][k],
            CONSTANT_NAME[MethodAccess[i][j][k]]}) 
            source=MethodSource[i][j][k]
            if source!=i then
               printf(fn,"  (%s)",{ClassName[source]})   
            end if 
            puts(fn,"\n")
         end for
      end for
   end for  
   puts(fn,"\n\n--- REGISTERED TYPES --- \n\n")
   for i=1 to length(TypeName) do
      puts(fn,TypeName[i]&"\n")
   end for
   printf(fn,"\n\n--- END OF LISTING FOR %02d/%02d/%04d %02d:%02d:%02d "
   &"--\n\n\n",{d[2],d[3],d[1]+1900,d[4],d[5],d[6]})
   close(fn)
end procedure

DEBUG_FILE=routine_id("debug_file")   

-- Program stack:
procedure push_stack(object caller,object thisObj,object thisMethod)
   ProgramCounter+=1
   if ProgramCounter>length(Caller) then 
      Caller&=STACK_EXPAND
      This&=STACK_EXPAND
      ThisMethod&=STACK_EXPAND
      CurrentException&=STACK_EXPAND
   end if
   Caller[ProgramCounter]=caller
   This[ProgramCounter]=thisObj
   ThisMethod[ProgramCounter]=thisMethod 
   CurrentException[ProgramCounter]=NIL
end procedure

procedure pop_stack() 
   integer ce
   if ProgramCounter>1 then
      ce=CurrentException[ProgramCounter]
      if ce>NIL then  
         if CurrentException[ProgramCounter-1]>NIL then
            fatal_error("Attempted to return from a method with a pending "
            &"exception to a method with a pending exception.")      
         end if 
         CurrentException[ProgramCounter-1]=ce
      end if      
   end if
   Caller[ProgramCounter]=NIL
   This[ProgramCounter]=NIL
   ThisMethod[ProgramCounter]=NIL 
   CurrentException[ProgramCounter]=NIL    
   ProgramCounter-=1
end procedure

-- The identifier type:
global type identifier(object x)
   if atom(x) then return FALSE end if
   if length(x)=0 then return FALSE end if
   if not find(x[1],ID_INIT) then return FALSE end if
   for i=2 to length(x) do
      if not find(x[i],ID_CHAR) then return FALSE end if
   end for
   return TRUE
end type   

type dot_identifier(object x)
   integer index    
   if atom(x) then return FALSE end if
   index=find('.',x) 
   if index=NIL then return identifier(x) end if
   if not identifier(x[1..index-1]) then return FALSE end if
   return dot_identifier(x[index+1..$])
end type

-- The entity types: 
type any_entity(object x)
   object xi                
   if atom(x) then return FALSE end if
   if length(x)!=3 then return FALSE end if
   if not equal(x[1],MARKER) then return FALSE end if
   xi=x[3] 
   if not integer(xi) then return FALSE end if
   if xi<1 then return FALSE end if
   if xi>ClassCounter then return FALSE end if
   xi=x[2]
   if not integer(xi) then return FALSE end if
   if xi=CLASS_FLAG then return TRUE end if   
   if xi<CLASS_FLAG then return FALSE end if     
   return xi<=InstanceCounter 
end type                                              

global type entity(object x)     
   integer x2
   if not any_entity(x) then return FALSE end if
   x2=x[2]
   if x2=CLASS_FLAG then
      return TRUE
   else
      return sequence(Instances[x2])
   end if
end type

global type instance_entity(object x) 
   integer x2
   if not any_entity(x) then return FALSE end if
   x2=x[2]
   if x2=CLASS_FLAG then
      return FALSE
   else
      return sequence(Instances[x2])   
   end if
end type    

global type class_entity(object x)
   if not any_entity(x) then return FALSE end if
   return x[2]=CLASS_FLAG   
end type                 

global type deleted_entity(object x)    
   integer x2
   if not any_entity(x) then return FALSE end if
   x2=x[2]
   if x2=CLASS_FLAG then
      return FALSE
   else
      return atom(Instances[x2])
   end if   
end type


-- Auxilliary types:  
global function make_sequence(object x) 
   object xi 
   if atom(x) then return {x} end if
   if length(x)!=3 then return x end if
   if not equal(x[1],MARKER) then return x end if
   xi=x[3] 
   if not integer(xi) then return x end if
   if xi<1 then return x end if
   if xi>ClassCounter then return x end if
   xi=x[2]
   if not integer(xi) then return x end if
   if xi<CLASS_FLAG then return x end if        
   if xi>InstanceCounter then return x end if
   return {x}   
end function   

function make_special_sequence(object x)
   if dot_identifier(x) then
      return {x}
   else
      return make_sequence(x)
   end if
end function

function resolve_name(object x,integer isType)
   integer index
   sequence names,pack
   if not dot_identifier(x) then return NIL end if
   if isType then names=TypeName else names=ClassName end if  
   index=find(x,names)
   if index!=NIL then return index end if
   if Context=METHOD then 
      pack="."&ClassName[Caller[ProgramCounter]]
      for i=length(pack) to 1 by -1 do
         if pack[i]='.' then
            pack=pack[2..i]
            exit   
         end if
      end for
   else
      pack=CurrentPackage&"."   
   end if
   for i=0 to length(PACKAGES) do
      if i!=0 then pack=PACKAGES[i]&"." end if
      for j=length(pack) to 2 by -1 do
         if pack[j]='.' then
            index=find(pack[1..j]&x,names)
            if index!=NIL then return index end if
         end if
      end for                       
   end for
   return NIL
end function   

function valid_class(object x,integer ct)
   integer super
   if not class_entity(x) then
      x=resolve_name(x,FALSE)
      if x!=NIL then
         x=CLASS_HEAD&x
      else
         return NIL
      end if   
   end if       
   if ct=0 then return x end if 
   super=SuperclassList[x[3]][$] 
   if super=ct then return x end if
   return super   
end function  

function valid_target(object x)
   integer super
   if not entity(x) then
      x=resolve_name(x,FALSE)
      if x!=NIL then
         x=CLASS_HEAD&x
      else
         return NIL
      end if   
   end if        
   super=SuperclassList[x[3]][$] 
   if super=ENTITY_ID then return x end if
   return super      
end function  

                               
function wrap_param(object x,integer paramCount) 
   integer len,parray                  
   sequence ret
   x=make_sequence(x) 
   len=length(x)
   if paramCount<0 then
      paramCount=-paramCount-1
      parray=TRUE  
   else
      parray=FALSE     
   end if
   if len<paramCount then  
      ret=x&repeat(NOTHING,paramCount-len)
      if parray then ret=append(ret,NONE) end if
   elsif len>=paramCount then 
      ret=x[1..paramCount]                 
      if parray then ret=append(ret,x[paramCount+1..len]) end if
   end if
   return ret    
end function    

function write_spec(object x)
   integer index,len
   object xi
   if atom(x) then x={x} end if 
   len=length(x)
   if len=0 or len>3 then return NIL end if 
   xi=x[1]
   if not find(xi,VALID_ACCESS) then return NIL end if
   if len=1 then return x end if
   if xi=PRIVATE then return NIL end if
   xi=resolve_name(x[2],TRUE)
   if xi=NIL then return NIL end if 
   x[2]=TypeName[xi]
   if len=2 then return x end if
   index=find(x[2],ENTITY_TYPES) 
   if index=NIL then return NIL end if 
   xi=valid_class(x[3],ENTITY_ID)
   if atom(xi) then return NIL end if
   return x
end function


-- Class information:
global function get_class(object x)
   integer index
   if entity(x) then
      x[2]=CLASS_FLAG        
      return x
   else
      index=resolve_name(x,FALSE)
      if index!=NIL then
         return CLASS_HEAD&index   
      end if
   end if
   return NOTHING
end function  

global function class_name(object x)       
   x=valid_class(x,NIL)
   if atom(x) then return NONE end if
   return ClassName[x[3]]
end function  


global function instance_of(object x,object y)
   if not instance_entity(x) then return FALSE end if
   y=valid_class(y,ENTITY_ID)
   if atom(y) then return FALSE end if
   return find(y[3],SuperclassList[x[3]])!=NIL                    
end function

global function extends(object x,object y)  
   x=valid_class(x,NIL)
   if atom(x) then return FALSE end if
   y=valid_class(y,NIL)
   if atom(y) then return FALSE end if
   return find(y[3],SuperclassList[x[3]])!=NIL    
end function

global function implements(object x,object y)    
   x=valid_class(x,ENTITY_ID)
   if atom(x) then return FALSE end if  
   y=valid_class(y,INTERFACE_ID)
   if atom(y) then return FALSE end if
   return find(y[3],InterfaceList[x[3]])!=NIL   
end function  

-- Exception handling:
global procedure throw(object target)
   if Context!=METHOD then
      fatal_error("throw()\nNot allowed.")   
   end if   
   if atom(valid_class(target,EXCEPTION_ID)) then
      fatal_error("throw()\nTarget is not an exception.")
   end if
   if CurrentException[ProgramCounter]>NIL then
      fatal_error("throw()\nException "&ClassName[target[3]]&" thrown.")   
   end if
   CurrentException[ProgramCounter]=target[3]
end procedure 

THROW=routine_id("throw")

global function catch(object target)
   integer ce,targetException
   if Context!=METHOD then
      fatal_error("catch()\nNot allowed.")   
   end if
   if atom(valid_class(target,EXCEPTION_ID)) then
      fatal_error("catch()\nTarget is not an exception.")
   end if           
   targetException=target[3] 
   ce=CurrentException[ProgramCounter]
   if ce<=NIL then return FALSE end if
   if not find(targetException,SuperclassList[ce]) then
      return FALSE   
   end if          
   CurrentException[ProgramCounter]=-ce
   return TRUE
end function    

global function caught() 
   integer ce
   if Context!=METHOD then
      fatal_error("caught()\nNot allowed.")   
   end if 
   ce=CurrentException[ProgramCounter]
   if ce>=NIL then return NOTHING end if
   return CLASS_HEAD&-ce
end function   

global procedure rethrow() 
   integer ce
   if Context!=METHOD then
      fatal_error("rethrow\nNot allowed.")   
   end if    
   ce=CurrentException[ProgramCounter]
   if ce>=NIL then return end if
   CurrentException[ProgramCounter]=-ce 
end procedure

global function success()
   if Context!=METHOD then
      fatal_error("success()\nNot allowed.")   
   end if
   return CurrentException[ProgramCounter]<=NIL      
end function   

global function failure()
   if Context!=METHOD then
      fatal_error("failure()\nNot allowed.")   
   end if
   return CurrentException[ProgramCounter]>NIL      
end function   
               
global function pending() 
   integer ce
   if Context!=METHOD then
      fatal_error("pending()\nNot allowed.")   
   end if    
   ce=CurrentException[ProgramCounter]                                 
   if ce<=NIL then return NOTHING end if
   return CLASS_HEAD&ce
end function


-- Class definition: 
global procedure package(object name)  
   if Context!=INITIALIZATION then
      fatal_error("package()\nNot allowed.")   
   end if
   if not identifier(name) then
      fatal_error("package()\nPackage name must be an identifier.")   
   end if  
   if length(CurrentPackage)=0 then
      CurrentPackage=name
   else
      CurrentPackage&='.'&name   
   end if
end procedure

global procedure end_package()  
   integer index,lastIndex,len 
   if Context!=INITIALIZATION then
      fatal_error("end_package()\nNot allowed.")   
   end if
   len=length(CurrentPackage)     
   if len=0 then
      fatal_error("end_package()\nNo current package.")   
   end if
   index=0
   lastIndex=1 
   while TRUE do
      index=find('.',CurrentPackage[index+1..len])+index
      if index<=lastIndex then
         CurrentPackage=CurrentPackage[1..lastIndex-1]
         return
      end if
      lastIndex=index
   end while
end procedure   


procedure new_class(sequence name,integer superclass)
   ClassCounter+=1
   ClassName&={name}  
   SuperclassList&={ClassCounter&SuperclassList[superclass]}
   InterfaceList&={NONE} 
   PropertyName&={PropertyName[superclass]}
   PropertySource&={PropertySource[superclass]}  
   PropertyType&={PropertyType[superclass]}
   PropertyDefault&={PropertyDefault[superclass]} 
   MethodName&={MethodName[superclass]}
   MethodSource&={MethodSource[superclass]} 
   MethodAccess&={MethodAccess[superclass]} 
   MethodParamCount&={MethodParamCount[superclass]}
   MethodID&={MethodID[superclass]}
   MethodReturn&={MethodReturn[superclass]}
   ClassInstances&=NIL   
end procedure


global function class(object name,object superclass,
object interfaceList) 
   sequence superclassList,temp
   object thisClass  
   if Context!=INITIALIZATION then
      fatal_error("class()\nNot allowed.")   
   end if                                     
   if not identifier(name) then
      fatal_error("class()\nInvalid class name.")   
   end if 
   if length(CurrentPackage)>0 then      
      name=CurrentPackage&"."&name   
   end if
   if find(name,ClassName) then
      fatal_error("class()\nClass "&name
      &" duplicates existing class name.")   
   end if  
   thisClass=valid_class(superclass,ENTITY_ID)
   if atom(thisClass) then
      if thisClass=INTERFACE_ID then
         fatal_error("class()\nClass "&name&" may not have interface class "
         &ClassName[superclass[3]]&" as superclass.")
      elsif thisClass=EXCEPTION_ID then
         fatal_error("class()\nClass "&name&" may not have exception class "
         &ClassName[superclass[3]]&" as superclass.")   
      else
         fatal_error("class()\nClass "&name&" has invalid superclass.")
      end if  
   end if 
   superclass=thisClass[3]  
   interfaceList=make_special_sequence(interfaceList)   
   for i=1 to length(interfaceList)  do
      thisClass=valid_class(interfaceList[i],INTERFACE_ID)
      if atom(thisClass) then 
         if thisClass=ENTITY_ID then
            fatal_error("class()\nClass "&name
            &" may not have normal class "
            &ClassName[interfaceList[i][3]]&" in interface list.")     
         elsif thisClass=EXCEPTION_ID then 
            fatal_error("class()\nClass "&name
            &" may not have exception class "
            &ClassName[interfaceList[i][3]]&" in interface list.")     
         else   
            fatal_error("class()\nClass "&name
            &" has invalid interface list.")    
         end if
      end if  
      thisClass=thisClass[3]
      interfaceList[i]=thisClass 
      if find(thisClass,interfaceList[1..i-1]) then
         fatal_error("class()\nClass "&name&" has duplicate interfaces.")   
      end if
   end for  
   superclassList=SuperclassList[superclass]
   temp=NONE 
   for i=1 to length(superclassList) do
      temp&=InterfaceList[superclassList[i]]   
   end for
   for i=1 to length(interfaceList) do
      temp&=SuperclassList[interfaceList[i]]   
   end for
   interfaceList=NONE
   for i=1 to length(temp) do     
      thisClass=temp[i]
      if not find(thisClass,temp[i+1..$]) then
         interfaceList&=thisClass   
      end if
   end for
   new_class(name,superclass)
   InterfaceList[ClassCounter]=interfaceList
   Context=CLASS_DEFINITION
   Caller[3]=ClassCounter
   return CLASS_HEAD&ClassCounter
end function  

global procedure end_class() 
   integer index,thisInterface
   sequence mname
   if Context!=CLASS_DEFINITION then
      fatal_error("end_class()\nNot allowed.")   
   end if
   for i=1 to length(InterfaceList[ClassCounter]) do
      thisInterface=InterfaceList[ClassCounter][i]
      for j=INSTANCE to CLASS do
         for k=1 to length(MethodName[thisInterface][j]) do
            mname=MethodName[thisInterface][j][k] 
            index=find(mname,MethodName[ClassCounter][j])
            if index=NIL or MethodAccess[ClassCounter][j][index]!=PUBLIC then
               fatal_error("end_class()\nDoes not implement "
               &ClassName[thisInterface]&" "&CONSTANT_NAME[j]
               &" method "&mname&" as a public method.")                 
            end if
         end for
      end for
   end for 
   ClassInstances[ClassCounter]=PropertyDefault[ClassCounter][CLASS]    
   Context=INITIALIZATION                                              
   Caller[3]=NIL
end procedure 

function merge_interfaces(sequence supers) 
   integer index,thisSuper
   sequence methodList,methods,name,temp  
   for i=1 to length(supers) do
      supers[i]=SuperclassList[supers[i]]   
   end for 
   temp=NONE
   for i=1 to length(supers) do
      temp&=supers[i]      
   end for   
   supers=NONE
   for i=1 to length(temp) do
      thisSuper=temp[i]
      if not find(thisSuper,temp[i+1..$]) then
         supers&=thisSuper   
      end if
   end for  
   methodList=MethodName[ClassCounter]            
   for i=1 to length(supers) do
      thisSuper=supers[i]
      for j=INSTANCE to CLASS do 
         methods=MethodName[thisSuper][j]
         for k=1 to length(methods) do
            name=methods[k]
            index=find(name,methodList[j])
            if index=NIL then    
               methodList[j]&={name}
               MethodName[ClassCounter][j]&={name}
               MethodSource[ClassCounter][j]&=MethodSource[thisSuper][j][k]
               MethodAccess[ClassCounter][j]&=PUBLIC
               MethodParamCount[ClassCounter][j]&=0
               MethodID[ClassCounter][j]&={{NOTHING}}   
            end if
         end for     
      end for
   end for
   return supers   
end function

global function interface(object name,object superclasses,
object instanceMethods,object classMethods) 
   integer index                    
   sequence methods
   object thisMeth,thisSuper
   if Context!=INITIALIZATION then
      fatal_error("interface()\nNot allowed.")   
   end if                                     
   if not identifier(name) then
      fatal_error("interface()\nInvalid interface name.")   
   end if
   if length(CurrentPackage)>0 then
      name=CurrentPackage&"."&name   
   end if
   if find(name,ClassName) then
      fatal_error("interface()\nInterface "&name
      &" duplicates existing class name.")   
   end if
   superclasses=make_special_sequence(superclasses)
   for i=1 to length(superclasses) do
      thisSuper=valid_class(superclasses[i],INTERFACE_ID)
      if atom(thisSuper) then     
         if thisSuper=ENTITY_ID then
            fatal_error("interface()\nClass "&name
            &" may not have normal class "
            &ClassName[superclasses[i][3]]&" as superclass.")
         elsif thisSuper=EXCEPTION_ID then
            fatal_error("interface()\nClass "&name
            &" may not have exception class "&ClassName[superclasses[i][3]]
            &" as superclass.")
         else
            fatal_error("interface()\nInterface "&name
            &" has an invalid superclass.")
         end if  
      end if   
      thisSuper=thisSuper[3]                          
      if find(thisSuper,superclasses[1..i-1]) then
         fatal_error("interface()\nInterface "&name
         &" has duplicate superclasses.")            
      end if
      superclasses[i]=thisSuper
   end for
   instanceMethods=make_special_sequence(instanceMethods)                         
   for i=1 to length(instanceMethods) do
      thisMeth=instanceMethods[i]
      if not identifier(thisMeth) then
         fatal_error("interface()\nInterface "&name
         &" has invalid instance method list.")   
      end if
      if find(thisMeth,instanceMethods[1..i-1]) then
         fatal_error("interface()\nInterface "&name
         &" has duplicate instance methods.")
      end if
   end for
   classMethods=make_special_sequence(classMethods)
   for i=1 to length(classMethods) do
      thisMeth=classMethods[i]
      if not identifier(thisMeth) then
         fatal_error("interface()\nInterface "&name
         &" has invalid class method list.")   
      end if
      if find(thisMeth,classMethods[1..i-1]) then
         fatal_error("interface()\nInterface "&name
         &" has duplicate class methods.")
      end if
   end for
   new_class(name,INTERFACE_ID)                         
   SuperclassList[ClassCounter]=
   ClassCounter&merge_interfaces(superclasses&INTERFACE_ID)
   methods=MethodName[ClassCounter][INSTANCE]
   for i=1 to length(instanceMethods) do
      thisMeth=instanceMethods[i]
      index=find(thisMeth,methods)
      if index!=NIL then
         MethodSource[ClassCounter][INSTANCE][index]=ClassCounter
      else
         MethodName[ClassCounter][INSTANCE]&={thisMeth}
         MethodSource[ClassCounter][INSTANCE]&=ClassCounter
         MethodAccess[ClassCounter][INSTANCE]&=PUBLIC
         MethodParamCount[ClassCounter][INSTANCE]&=0
         MethodID[ClassCounter][INSTANCE]&=BAD_METHOD
         MethodReturn[ClassCounter][INSTANCE]&={NOTHING}   
      end if
   end for 
   methods=MethodName[ClassCounter][CLASS]
   for i=1 to length(classMethods) do
      thisMeth=classMethods[i]
      index=find(thisMeth,methods)
      if index!=NIL then 
         MethodSource[ClassCounter][CLASS][index]=ClassCounter
      else
         MethodName[ClassCounter][CLASS]&={thisMeth} 
         MethodSource[ClassCounter][CLASS]&=ClassCounter
         MethodAccess[ClassCounter][CLASS]&=PUBLIC
         MethodParamCount[ClassCounter][CLASS]&=0
         MethodID[ClassCounter][CLASS]&=BAD_METHOD
         MethodReturn[ClassCounter][CLASS]&={NOTHING}   
      end if 
   end for
   return CLASS_HEAD&ClassCounter
end function

global function exception(object name,object superclass) 
   object thisClass  
   if Context!=INITIALIZATION then
      fatal_error("exception()\nNot allowed.")   
   end if                                     
   if not identifier(name) then
      fatal_error("exception()\nInvalid exception name.")   
   end if
   if length(CurrentPackage)>0 then
      name=CurrentPackage&"."&name   
   end if
   if find(name,ClassName) then
      fatal_error("exception()\nClass "&name
      &" duplicates existing class name.")   
   end if          
   thisClass=valid_class(superclass,EXCEPTION_ID)
   if atom(thisClass) then
      if thisClass=ENTITY_ID then
         fatal_error("exception()\nClass "&name&" may not have normal class "
         &ClassName[superclass[3]]&" as superclass.")
      elsif thisClass=INTERFACE_ID then
         fatal_error("exception()\nClass "&name
         &" may not have interface class "
         &ClassName[superclass[3]]&" as superclass.")
      else
         fatal_error("exception()\nClass "&name&" has invalid superclass.")
      end if
   end if
   superclass=thisClass[3]
   new_class(name,superclass)
   return CLASS_HEAD&ClassCounter
end function   

procedure initialize_program() 
   integer ID,index  
   sequence params,thisClass
   ProgramCounter=0
   Context=METHOD
   for i=18 to ClassCounter do
      index=find("initialize",MethodName[i][CLASS])
      LastInitialize=i
      if index!=NIL then
         ID=MethodID[i][CLASS][index]
         if ID!=BAD_METHOD then
            thisClass=CLASS_HEAD&i 
            params=wrap_param(NONE,MethodParamCount[i][CLASS][index]) 
            push_stack(MethodSource[i][CLASS][index],thisClass,
            "initialize")  
            if ID<0 then
               call_proc(-ID-2,params)
            else
               VOID=call_func(ID,params)    
            end if
            CurrentException[ProgramCounter]=NIL 
            pop_stack()         
         end if
      end if
   end for 
end procedure

procedure terminate_program() 
   integer ID,index 
   sequence params,thisClass            
   if TerminationClass=NIL then 
      TerminationClass=LastInitialize
   else
      TerminationClass-=1      
   end if
   for i=TerminationClass to 18 by -1 do
      TerminationClass=i 
      index=find("terminate",MethodName[i][CLASS])
      if index!=NIL then
         ID=MethodID[i][CLASS][index]
         if ID!=BAD_METHOD then
            thisClass=CLASS_HEAD&i 
            params=wrap_param(NONE,MethodParamCount[i][CLASS][index]) 
            push_stack(MethodSource[i][CLASS][index],thisClass,
            "terminate")    
            if ID<0 then
               call_proc(-ID-2,params)   
            else
               VOID=call_func(ID,params)
            end if
            CurrentException[ProgramCounter]=NIL
            pop_stack()          
         end if
      end if
   end for   
end procedure

global procedure main_program(object target)    
   integer ID,index,targetClass     
   object params
   if Context!=INITIALIZATION then
      fatal_error("main_program()\nNot allowed.")   
   end if
   target=valid_class(target,ENTITY_ID)                                        
   if atom(target) then
      fatal_error("main_program()\nTarget is not a class.")   
   end if 
   targetClass=target[3]                                                 
   index=find("main",MethodName[targetClass][CLASS])
   if index=NIL then
      fatal_error("main_program()\nClass "&ClassName[targetClass]
      &" does not define class method main.")     
   end if  
   initialize_program()                                  
   ID=MethodID[targetClass][CLASS][index]
   ProgramCounter=0    
   Context=METHOD  
   if ID!=BAD_METHOD then
      params=wrap_param(command_line(),
      MethodParamCount[targetClass][CLASS][index])
      push_stack(targetClass,target,"main")
      if ID<0 then
         call_proc(-ID-2,params)
      else
         VOID=call_func(ID,params)   
      end if 
      CurrentException[ProgramCounter]=NIL
      pop_stack()    
   end if
   terminate_program() 
   abort(0) 
end procedure   

global procedure start_program()  
   if Context!=INITIALIZATION then
      fatal_error("start_program()\nNot allowed.")   
   end if                                        
   initialize_program() 
   ProgramCounter=0
   Context=METHOD
   push_stack(PROGRAM[3],PROGRAM,"main")   
end procedure    

global procedure end_program() 
   if Context!=METHOD then
      fatal_error("end_program()\nNot allowed.")   
   end if   
   CurrentException[ProgramCounter]=NIL                                     
   pop_stack()  
   terminate_program()
   abort(0)   
end procedure

sequence merr merr="method"  

type good_rid(object x)  
   integer len
   if integer(x) then return x>BAD_METHOD end if
   if atom(x) then return FALSE end if
   len=length(x)
   if len<1 or len>2 then return FALSE end if         
   x=x[1]
   if integer(x) then return x>BAD_METHOD else return FALSE end if
end type
 
global procedure method(object name,object targetType,object access,
object paramCount,object ID) 
   integer index,mAccess,source 
   object ret
   if Context!=CLASS_DEFINITION then
      fatal_error(merr&"()\nNot allowed.")   
   end if                                  
   if not identifier(name) then
      fatal_error(merr&"()\nInvalid method name.")
   end if                                          
   if not find(targetType,VALID_TARGET_TYPE) then
      fatal_error(merr&"()\nMethod "&name&" has invalid target type.")   
   end if                                                              
   if not find(access,VALID_ACCESS) then
      fatal_error(merr&"()\n"&CONSTANT_CAP_NAME[targetType]&" method "
      &name&" has invalid access.")
   end if 
   if not integer(paramCount) then
      fatal_error(merr&"()\n"&CONSTANT_CAP_NAME[targetType]&" method "   
      &name&" has invalid parameter count.")
   end if   
   if not good_rid(ID) then
      fatal_error(merr&"()\n"&CONSTANT_CAP_NAME[targetType]&" method "
      &name&" has invalid routine id.")      
   end if
   if integer(ID) then
      ret=NIL
   else
      if length(ID)=1 then
         ret=NOTHING
      else
         ret=ID[2]   
      end if      
      ID=-ID[1]-2
   end if
   index=find(name,MethodName[ClassCounter][targetType])
   if index=NIL then
      MethodName[ClassCounter][targetType]&={name}
      MethodSource[ClassCounter][targetType]&=ClassCounter
      MethodAccess[ClassCounter][targetType]&=access
      MethodParamCount[ClassCounter][targetType]&=paramCount
      MethodID[ClassCounter][targetType]&=ID
      MethodReturn[ClassCounter][targetType]&={ret}
   else 
      source=MethodSource[ClassCounter][targetType][index]                   
      if source=ClassCounter then
         fatal_error(merr&"()\n"&CONSTANT_CAP_NAME[targetType]
         &" method "&name&" already defined.")    
      end if
      mAccess=MethodAccess[ClassCounter][targetType][index]
      if access>mAccess then
         fatal_error(merr&"()\n"&CONSTANT_CAP_NAME[targetType]
         &" method "&name&" has more restrictive access than superclass "
         &ClassName[source]&" method.")
      end if
      MethodSource[ClassCounter][targetType][index]=ClassCounter 
      MethodAccess[ClassCounter][targetType][index]=access 
      MethodParamCount[ClassCounter][targetType][index]=paramCount
      MethodID[ClassCounter][targetType][index]=ID 
      MethodReturn[ClassCounter][targetType][index]=ret   
   end if
end procedure                                         

global procedure null_method(object name,object targetType,object access,
object ret) 
   merr="null_method"
   method(name,targetType,access,0,{NULL_METHOD_ID,ret}) 
   merr="method"
end procedure 

global procedure super_method(object name,object targetType)   
   merr="super_method"
   method(name,targetType,PUBLIC,-1,SUPER_METHOD_ID) 
   merr="method"           
end procedure 

function get_x()             
   sequence name
   name=ThisMethod[ProgramCounter]
   return call_func(GET_PROPERTY,{This[ProgramCounter],name[5..$]})   
end function  
constant GET_X=routine_id("get_x") 
  
procedure set_x(object newValue)             
   integer index,len,targetType,thisClass
   sequence name,propType,thisEntity
   name=ThisMethod[ProgramCounter]
   name=name[5..$]       
   thisEntity=This[ProgramCounter]
   thisClass=thisEntity[3]
   if thisEntity[2]!=CLASS_FLAG then
      targetType=INSTANCE
   else
      targetType=CLASS      
   end if             
   index=find(name,PropertyName[thisClass][targetType])
   propType=PropertyType[thisClass][targetType][index]
   len=length(propType)
   if len=1 then
      call_proc(VALIDATE,{newValue,propType[1],REQUIRED})
   elsif len=2 then
      call_proc(VALIDATE,{newValue,propType,REQUIRED})
   end if   
   if CurrentException[ProgramCounter]<=NIL then
      call_proc(SET_PROPERTY,{This[ProgramCounter],name,newValue})    
   end if   
end procedure  
constant SET_X={routine_id("set_x"),NOTHING}  

sequence perr perr="property"

global procedure property(object name,object targetType,object readAccess,
object writeAccess,object default)
   integer index,source
   if Context!=CLASS_DEFINITION then
      fatal_error(perr&"()\nNot allowed.")   
   end if                                    
   if not identifier(name) then
      fatal_error(perr&"()\nInvalid property name.")   
   end if       
   if not find(targetType,VALID_TARGET_TYPE) then
      fatal_error(perr&"()\nProperty "&name&" has invalid target type.")   
   end if
   if not find(readAccess,VALID_ACCESS) then
      fatal_error(perr&"()\nProperty "&name&" has invalid read access.")
   end if                                                                  
   writeAccess=write_spec(writeAccess)
   if atom(writeAccess) then
      fatal_error(perr&"()\nProperty "&name&" has invalid write access.")
   end if                                                                  
   index=find(name,PropertyName[ClassCounter][targetType])
   if index!=NIL then
      source=PropertySource[ClassCounter][targetType][index]
      if source=ClassCounter then
         fatal_error(perr&"()\n"&CONSTANT_CAP_NAME[targetType]
         &" property "&name&" already defined.")   
      end if
      fatal_error(perr&"()\n"&CONSTANT_CAP_NAME[targetType]&" property "
      &name&" attempts to override "&ClassName[source]&" property.")
   end if
   PropertyName[ClassCounter][targetType]&={name}
   PropertySource[ClassCounter][targetType]&=ClassCounter 
   PropertyType[ClassCounter][targetType]&=
   {writeAccess[2..$]}
   writeAccess=writeAccess[1]
   PropertyDefault[ClassCounter][targetType]&={default}
   merr=perr
   if readAccess!=PRIVATE then 
      method("get_"&name,targetType,readAccess,0,GET_X)   
   end if                                               
   if writeAccess!=PRIVATE then
      method("set_"&name,targetType,writeAccess,1,SET_X)   
   end if        
   merr="method"
end procedure 

function x_event(sequence params)
   sequence evt
   object handler,handlers,ret  
   evt=ThisMethod[ProgramCounter]
   handlers=call_func(GET_PROPERTY,{This[ProgramCounter],evt}) 
   params={This[ProgramCounter],evt}&params
   ret=NOTHING
   for i=1 to length(handlers) do
      handler=handlers[i]
      if instance_of(handler,METHOD_WRAPPER) then
         ret=call_func(CALL_METHOD,{handler,"call",params})
      end if
      if failure() or not equal(ret,NOTHING) then exit end if
   end for                                                  
   return ret    
end function   
constant X_EVENT=routine_id("x_event")

function set_x_event(object handlers)
   sequence evt 
   if equal(handlers,NOTHING) then
      handlers=NONE   
   elsif atom(handlers) or entity(handlers) then
      handlers={handlers}   
   end if  
   for i=1 to length(handlers) do
      if not instance_of(handlers[i],METHOD_WRAPPER) then
         call_proc(THROW,{TYPE_CHECK_FAILURE})
         return NOTHING   
      end if      
   end for  
   evt=ThisMethod[ProgramCounter] 
   call_proc(SET_PROPERTY,{This[ProgramCounter],evt[5..$],handlers})
   return NOTHING   
end function     
constant SET_X_EVENT=routine_id("set_x_event")

global procedure event(object name,object linkAccess,object raiseAccess)
   perr="event"
   property(name,INSTANCE,linkAccess,PRIVATE,NONE)
   perr="property"
   merr="event" 
   method("set_"&name,INSTANCE,linkAccess,1,SET_X_EVENT)
   method(name,INSTANCE,raiseAccess,-1,X_EVENT) 
   merr="method"  
end procedure         

-- Kernel exceptions:  
package("ME_Kernel") 


global constant Calling_Error=exception("Calling_Error",Exception)


global constant Bad_Parameter=exception("Bad_Parameter",Calling_Error)


global constant Missing_Parameter=
exception("Missing_Parameter",Bad_Parameter) 


global constant Type_Check_Failure=
exception("Type_Check_Failure",Bad_Parameter)
TYPE_CHECK_FAILURE=Type_Check_Failure  


global constant Invalid_Type=exception("Invalid_Type",Bad_Parameter)  
      
      
global constant Invalid_Target=exception("Invalid_Target",Calling_Error)
      
      
global constant Undefined_Method=exception("Undefined_Method",Calling_Error)


global constant Undefined_Property=
exception("Undefined_Property",Calling_Error)


global constant Access_Denied=exception("Access_Denied",Calling_Error)


global constant Bad_Serialization=
exception("Bad_Serialization",Exception)


global constant Serialize_Error=
exception("Serialize_Error",Bad_Serialization)
SERIALIZE_ERROR_ID=Serialize_Error[3]


global constant Deserialize_Error=
exception("Deserialize_Error",Bad_Serialization) 
DESERIALIZE_ERROR_ID=Deserialize_Error[3]


-- Class operations:
global function this()
   if Context!=METHOD then
      fatal_error("this()\nNot allowed.")
   end if
   return This[ProgramCounter]
end function   

global function this_class() 
   sequence thisEntity
   if Context!=METHOD then
      fatal_error("this_class()\nNot allowed.")
   end if
   thisEntity=This[ProgramCounter]
   thisEntity[2]=CLASS_FLAG
   return thisEntity    
end function 

global function this_method()
   if Context!=METHOD then
      fatal_error("this_method()\nNot allowed.")
   end if
   return ThisMethod[ProgramCounter]   
end function

global function last()
   if Context!=METHOD then
      fatal_error("last()\nNot allowed.")
   end if              
   if ProgramCounter=1 then return NOTHING end if
   return This[ProgramCounter-1]
end function   

global function last_class() 
   sequence lastEntity
   if Context!=METHOD then
      fatal_error("last_class()\nNot allowed.")
   end if          
   if ProgramCounter=1 then return NOTHING end if
   lastEntity=This[ProgramCounter-1]
   lastEntity[2]=CLASS_FLAG
   return lastEntity    
end function 

global function last_method()
   if Context!=METHOD then
      fatal_error("last_method()\nNot allowed.")
   end if              
   if ProgramCounter=1 then return NONE end if
   return ThisMethod[ProgramCounter-1]
end function   
                                 
global function get_property(object target,object name)
   integer index,targetClass,targetInstance,targetType
   if Context!=METHOD then
      fatal_error("get_property()\nNot allowed.")   
   end if                                        
   target=valid_target(target) 
   if atom(target) then
      throw(Invalid_Target)
      return NOTHING    
   end if 
   targetClass=target[3]
   targetInstance=target[2]
   if targetInstance!=CLASS_FLAG then
      targetType=INSTANCE
   else
      targetType=CLASS   
   end if
   index=find(name,PropertyName[targetClass][targetType])
   if index=NIL then  
      throw(Undefined_Property)
      return NOTHING
   end if
   if PropertySource[targetClass][targetType][index]!=
   Caller[ProgramCounter] then
      throw(Access_Denied)
      return NOTHING           
   end if                                                            
   if targetType=INSTANCE then
      return Instances[targetInstance][index]
   else
      return ClassInstances[targetClass][index]   
   end if
end function
GET_PROPERTY=routine_id("get_property")

global procedure set_property(object target,object name,object val) 
   integer index,targetClass,targetInstance,targetType
   if Context!=METHOD then
      fatal_error("set_property()\nNot allowed.")   
   end if                                        
   target=valid_target(target) 
   if atom(target) then
      throw(Invalid_Target)
      return
   end if 
   targetClass=target[3]
   targetInstance=target[2]
   if targetInstance!=CLASS_FLAG then
      targetType=INSTANCE
   else 
      targetType=CLASS
   end if                  
   index=find(name,PropertyName[targetClass][targetType])
   if index=NIL then  
      throw(Undefined_Property)
      return
   end if
   if PropertySource[targetClass][targetType][index]!=
   Caller[ProgramCounter] then
      throw(Access_Denied)
      return           
   end if
   if targetType=INSTANCE then
      Instances[targetInstance][index]=val
   else
      ClassInstances[targetClass][index]=val   
   end if
end procedure  
SET_PROPERTY=routine_id("set_property")  

function undefined_method(sequence target,integer targetClass,
integer targetType,object name,object params)
   integer index,pc,source
   object ID,ret
   index=find("undefined_method",MethodName[targetClass][targetType])
   if index=NIL then return NIL end if
   source=MethodSource[targetClass][targetType][index]  
   pc=MethodParamCount[targetClass][targetType][index] 
   ID=MethodID[targetClass][targetType][index] 
   params=make_sequence(params)
   params=wrap_param({name}&params,pc)
   push_stack(source,target,"undefined_method")
   if ID<0 then
      ret=MethodReturn[targetClass][targetType][index] 
      call_proc(-ID-2,params)
   else
      ret=call_func(ID,params)    
   end if
   pop_stack() 
   return {ret}          
end function

global function call_method(object target,object name,object params)
   integer access,caller,ID,index,paramCount,source,targetClass
   integer targetInstance,targetType 
   object ret
   if Context!=METHOD then
      fatal_error("call_method()\nNot allowed.")   
   end if                                        
   target=valid_target(target) 
   if atom(target) then
      throw(Invalid_Target)
      return NOTHING    
   end if
   targetClass=target[3]
   targetInstance=target[2] 
   if targetInstance=1 then
      throw(Invalid_Target)
      return NOTHING
   elsif targetInstance!=CLASS_FLAG then
      targetType=INSTANCE
   else 
      targetType=CLASS
   end if   
   if find(name,NO_DIRECT_CALLS[targetType]) then               
      throw(Access_Denied) 
      return NOTHING
   end if
   index=find(name,MethodName[targetClass][targetType])
   if index=NIL then  
      ret=undefined_method(target,targetClass,targetType,name,params) 
      if atom(ret) then
         throw(Undefined_Method)
         return NOTHING
      end if 
      return ret[1]  
   end if 
   access=MethodAccess[targetClass][targetType][index] 
   source=MethodSource[targetClass][targetType][index]
   while TRUE do 
      if access=PUBLIC then exit end if 
      caller=Caller[ProgramCounter]
      if find(caller,SuperclassList[source]) then exit end if
      if access=PROTECTED then             
         if find(caller,SuperclassList[targetClass]) then exit end if
      end if 
      throw(Access_Denied)
      return NOTHING   
   end while 
   ID=MethodID[targetClass][targetType][index] 
   paramCount=MethodParamCount[targetClass][targetType][index]
   params=wrap_param(params,paramCount)
   push_stack(source,target,name)  
   if ID<0 then
      ret=MethodReturn[targetClass][targetType][index] 
      call_proc(-ID-2,params)   
   else
      ret=call_func(ID,params)
   end if
   pop_stack() 
   return ret         
end function
CALL_METHOD=routine_id("call_method")

global function call_super(object params)
   integer access,caller,ID,index,paramCount,source,targetClass
   integer targetType 
   sequence name,target
   object ret
   if Context!=METHOD then
      fatal_error("call_super()\nNot allowed.")   
   end if                    
   target=This[ProgramCounter]                    
   if target[2]!=CLASS_FLAG then
      targetType=INSTANCE 
      if atom(Instances[target[2]]) then
         throw(Invalid_Target)
         return NOTHING   
      end if
   else
      targetType=CLASS
   end if 
   caller=Caller[ProgramCounter]               
   targetClass=SuperclassList[caller][2] 
   name=ThisMethod[ProgramCounter]    
   if find(name,NO_SUPER_CALLS[targetType]) then
      throw(Access_Denied)
      return NOTHING   
   end if
   index=find(name,MethodName[targetClass][targetType])
   if index=NIL then
      ret=undefined_method(target,targetClass,targetType,name,params) 
      if atom(ret) then
         throw(Undefined_Method)
         return NOTHING
      end if 
      return ret[1]  
   end if
   access=MethodAccess[targetClass][targetType][index] 
   source=MethodSource[targetClass][targetType][index] 
   if access=PRIVATE then
      throw(Access_Denied)
      return NOTHING   
   end if  
   ID=MethodID[targetClass][targetType][index] 
   paramCount=MethodParamCount[targetClass][targetType][index]
   params=wrap_param(params,paramCount) 
   push_stack(source,target,name)
   if ID<0 then
      ret=MethodReturn[targetClass][targetType][index]
      call_proc(-ID-2,params)
   else
      ret=call_func(ID,params)   
   end if
   pop_stack()
   return ret
end function  

function super_method_implementation(sequence params)
   return call_super(params)   
end function                
SUPER_METHOD_ID=routine_id("super_method_implementation")

procedure null_method_implementation()
end procedure                        
NULL_METHOD_ID=routine_id("null_method_implementation")

-- Serialization:    
-- 0 - 246 literal integer value
-- 247 NOTHING
-- 248 2 byte integer
-- 249 3 byte integer
-- 250 4 byte integer
-- 251 4 byte atom
-- 252 8 byte atom
-- 253 string
-- 254 sequence
-- 255 entity next byte is type:
---- 0 pointer
---- 1 instance
---- 2 class

global type string(object x)
   object c
   if atom(x) then return FALSE end if
   for i=1 to length(x) do
      c=x[i]
      if not integer(c) then return FALSE end if
      if c<0 then return FALSE end if
      if c>255 then return FALSE end if
   end for                             
   return TRUE
end type   

type property_sequence(object x)  
   sequence temp
   object prop,x1,x2
   if atom(x) then return FALSE end if
   if length(x)!=2 then return FALSE end if
   temp=NONE
   x1=x[1]
   x2=x[2]
   if atom(x1) or atom(x2) or length(x1)!=length(x2) then 
      return FALSE
   end if
   for i=1 to length(x1) do
      prop=x1[i]
      if not identifier(prop) then return FALSE end if
      if find(prop,temp) then return FALSE end if
      temp&={prop}
   end for
   return TRUE
end type               

global function serialized()
   if Context!=METHOD then
      fatal_error("serialized()\nNot allowed.")   
   end if 
   return ThisSer
end function    

global procedure cancel_serialize()  
   if Context!=METHOD then
      fatal_error("cancel_serialize()\nNot allowed.")   
   end if 
   Instances[1]=NIL   
end procedure

procedure put_bc(sequence b) 
   if SerFile=-1 then
      Bytecode&=b
   else
      puts(SerFile,b)   
   end if
end procedure

procedure ser(object x)        
   integer ce,ID,flag,index,len,negative,superClass,targetClass
   integer targetInstance
   sequence supers   
   object props,temp
   if integer(x) then
      if x>=0 and x<=246 then 
         put_bc({x})
         return
      end if
      if x>0 then
         negative=FALSE
      else
         negative=TRUE
         x=-x-1
      end if
      poke4(MEM,x)
      temp=peek({MEM,4})
      for i=4 to 3 do
         if temp[i]=0 and temp[i-1]<128 then
            temp=temp[1..i-1]   
         end if
      end for                
      if negative then temp=255-temp end if 
      put_bc((246+length(temp))&temp)
   elsif atom(x) then  
      if x=NOTHING then
         put_bc({247})
      else
         temp=machine_func(48,x) -- atom_to_float32
         if x=machine_func(49,temp) then  -- float32_to_atom
            put_bc(251&temp)
         else
            put_bc(252&machine_func(46,x)) -- atom_to_float64
         end if                              
      end if
   elsif string(x) then
      put_bc({253})
      ser(length(x))
      put_bc(x)  
   elsif not entity(x) then
      len=length(x)
      put_bc({254})
      ser(len)
      for i=1 to len do
         ser(x[i])   
      end for 
   else
      targetClass=x[3] 
      targetInstance=x[2]
      if targetInstance=CLASS_FLAG then
         put_bc({255,CLASS})
         ser(ClassName[targetClass])
         return     
      end if
      props=Instances[targetInstance]
      if atom(props) then
         put_bc({247})
         return
      end if
      index=find(x,InstanceList)
      if index!=NIL then 
         put_bc({255,0})
         ser(index)
         return
      end if 
      ThisSer=x
      ThisSer[2]=1
      InstanceList=append(InstanceList,x)
      Instances[1]=props
      supers=SuperclassList[targetClass] 
      flag=TRUE   
      for i=1 to length(supers) do
         superClass=supers[i]   
         index=find("serialize",MethodName[superClass][INSTANCE])
         if index=NIL then exit end if
         if MethodSource[superClass][INSTANCE][index]=superClass then
            ID=MethodID[superClass][INSTANCE][index]
            if ID!=BAD_METHOD then      
               push_stack(MethodSource[superClass][INSTANCE][index],x,
               "serialize")
               if ID<0 then 
                  call_proc(-ID-2,NONE)   
               else
                  VOID=call_func(ID,NONE)  
               end if
               pop_stack()    
            end if 
            ce=CurrentException[ProgramCounter] 
            if ce>NIL then
               SerErr=ce
               CurrentException[ProgramCounter]=NIL  
               flag=FALSE
               exit 
            elsif atom(Instances[1]) then 
               ser(NOTHING)
               flag=FALSE
               exit  
            end if 
         end if
      end for        
      if flag then      
         put_bc({255,INSTANCE})
         ser(ClassName[targetClass])
         ser({PropertyName[targetClass][INSTANCE],Instances[1]}) 
      end if  
      ThisSer=NOTHING
      Instances[1]=NIL                                  
   end if
end procedure  

global function serialize(object x,object filename)
   integer toFile
   sequence bytes  
   if Context!=METHOD then
      fatal_error("serialize()\nNot allowed.")   
   end if 
   push_stack(ENTITY_ID,Entity,"serialize")       
   if SerInProgress then
      throw(Serialize_Error)               
      pop_stack()
      return NONE      
   end if 
   if not string(filename) or length(filename)=0 then
      toFile=FALSE 
      SerFile=-1 
      Bytecode=NONE
   else 
      toFile=TRUE  
      SerFile=open(filename,"wb")
      if SerFile=-1 then   
         throw(Serialize_Error)
         pop_stack()
         return NONE      
      end if  
   end if
   SerInProgress=TRUE 
   SerErr=NIL
   InstanceList=NONE 
   ser("ME")
   ser(CURRENT_VERSION)
   ser(x)
   SerInProgress=FALSE  
   if toFile=FALSE then
      bytes=Bytecode 
      Bytecode=NONE
      if SerErr!=NIL then
         throw(CLASS_HEAD&SerErr)
         pop_stack()
         return NONE 
      end if    
      pop_stack()
      return bytes
   end if          
   close(SerFile) 
   if SerErr=NIL then 
      pop_stack()
      return filename 
   end if
   if platform()=3 then
      system("rm "&filename,2)
   else
      system("del "&filename,2)   
   end if
   throw(CLASS_HEAD&SerErr)
   pop_stack()
   return NONE 
end function   

function get_bc(integer count)  
   integer lastc  
   sequence ret 
   if SerErr!=NIL then return NIL end if
   if count=0 then return NONE end if
   if SerFile=-1 then  
      lastc=DesPtr+count-1
      if lastc>length(Bytecode) then 
         SerErr=DESERIALIZE_ERROR_ID
         return NONE   
      else
         ret=Bytecode[DesPtr..lastc]
         DesPtr+=count 
         return ret
      end if       
   else
      ret=repeat(-1,count)
      for i=1 to count do
         ret[i]=getc(SerFile)   
      end for                
      if ret[count]=-1 then 
         SerErr=DESERIALIZE_ERROR_ID
         return NONE
      else
         return ret   
      end if
   end if
end function
 
function des()
   integer c,ce,iCtr,ID,index,index2,superClass
   atom ret
   sequence bytes,iPropNames,iPropValues,new,propNames,seq,supers
   object elem,name,pointer,props
   if SerErr then return NIL end if
   bytes=get_bc(1) 
   if SerErr then return NIL end if
   c=bytes[1]
   if c<=246 then
      return c
   elsif c=247 then
      if SVersion=1 then return c end if
      return NOTHING
   elsif c=248 then
      bytes=get_bc(2)
      if SerErr then return NIL end if
      if bytes[2]>127 then
         bytes&={255,255}
      else
         bytes&={0,0}
      end if
      poke(MEM,bytes)  
      ret=peek4s(MEM)
      if SVersion=1 and ret<0 then ret+=1 end if
      return ret
   elsif c=249 then
      bytes=get_bc(3) 
      if SerErr then return NIL end if
      if bytes[3]>127 then
         bytes&=255
      else
         bytes&=0
      end if
      poke(MEM,bytes)
      ret=peek4s(MEM)
      if SVersion=1 and ret<0 then ret+=1 end if
      return ret
   elsif c=250 then
      bytes=get_bc(4)
      if SerErr then return NIL end if
      poke(MEM,bytes) 
      ret=peek4s(MEM)
      if SVersion=1 and ret<0 then ret+=1 end if
      return ret
   elsif c=251 then
      bytes=get_bc(4) 
      if SerErr then return NIL end if
      return machine_func(49,bytes) -- float32_to_atom
   elsif c=252 then 
      bytes=get_bc(8) 
      if SerErr then return NIL end if
      return machine_func(47,bytes) -- float64_to_atom 
   elsif c=253 then
      elem=des()
      if SerErr then return NIL end if
      if not integer(elem) or elem<0 then
         SerErr=DESERIALIZE_ERROR_ID
         return NIL   
      end if
      bytes=get_bc(elem)
      if SerErr then return NIL end if
      return bytes
   elsif c=254 then
      elem=des()
      if SerErr then return NIL end if
      if not integer(elem) or elem<0 then
         SerErr=DESERIALIZE_ERROR_ID
         return NIL   
      end if
      seq=repeat(NIL,elem)
      for i=1 to elem do
         seq[i]=des()
         if SerErr then return NIL end if
      end for
      if equal(seq,OLD_NOTHING) and SVersion=1 then return NOTHING end if
      return seq
   else   
      bytes=get_bc(1) 
      if SerErr then return NIL end if
      c=bytes[1]  
      if c=0 then
         pointer=des()
         if SerErr then return NIL end if
         if not integer(pointer) or pointer<1 
         or pointer>length(InstanceList) then 
            SerErr=DESERIALIZE_ERROR_ID
            return NIL 
         end if        
         return InstanceList[pointer]    
      elsif c=INSTANCE then              
         name=des()  
         if SerErr then return NIL end if
         index=find(name,ClassName)
         if index=NIL then 
            SerErr=DESERIALIZE_ERROR_ID
            return NIL
         end if 
         iPropNames=PropertyName[index][INSTANCE]
         iPropValues=PropertyDefault[index][INSTANCE] 
         InstanceCounter+=1 
         if InstanceCounter>length(Instances) then
            Instances&=repeat(NIL,length(Instances))   
         end if   
         iCtr=InstanceCounter               
         new={MARKER,iCtr,index} 
         InstanceList=append(InstanceList,new)                          
         props=des() 
         if SerErr then return NIL end if
         if not property_sequence(props) then 
            SerErr=DESERIALIZE_ERROR_ID
            return NIL
         end if                                        
         propNames=props[1]
         for i=1 to length(propNames) do
            index2=find(propNames[i],iPropNames)   
            if index2!=NIL then
               iPropValues[index2]=props[2][i]   
            end if
         end for 
         Instances[iCtr]=iPropValues
         supers=SuperclassList[index]  
         for i=1 to length(supers) do    
            superClass=supers[i]
            index2=find("deserialize",MethodName[superClass][INSTANCE])
            if index2=NIL then exit end if
            if MethodSource[superClass][INSTANCE][index2]=superClass then
               ID=MethodID[superClass][INSTANCE][index2]
               if ID!=BAD_METHOD then
                  push_stack(superClass,new,"deserialize")
                  if ID<0 then
                     call_proc(-ID-2,NONE)
                  else
                     VOID=call_func(ID,NONE)   
                  end if 
                  pop_stack() 
                  ce=CurrentException[ProgramCounter]
                  if ce>NIL then           
                     SerErr=ce
                     CurrentException[ProgramCounter]=NIL
                     return NIL 
                  elsif atom(Instances[iCtr]) then  
                     InstanceList[$]=NOTHING
                     return NOTHING
                  end if
               end if
            end if        
         end for      
         return new
      elsif c=CLASS then  
         name=des() 
         if SerErr then return NIL end if
         index=find(name,ClassName)
         if index=NIL then 
            SerErr=DESERIALIZE_ERROR_ID
            return NIL
         end if       
         return CLASS_HEAD&index
      end if
   end if
end function  

procedure close_des()
   if SerFile=-1 then
      Bytecode=NONE
   else
      close(SerFile)   
   end if
end procedure

global function deserialize(object x)  
   integer index
   object thisInstance                     
   if Context!=METHOD then
      fatal_error("deserialize()\nNot allowed.")   
   end if                                      
   push_stack(ENTITY_ID,Entity,"deserialize")
   if SerInProgress or not string(x) or length(x)=0 then
      throw(Deserialize_Error) 
      pop_stack()
      return NOTHING   
   end if  
   if x[1]=253 then
      SerFile=-1 
      Bytecode=x 
      DesPtr=1  
   else
      SerFile=open(x,"rb")
      if SerFile=-1 then
         throw(Deserialize_Error)
         pop_stack()
         return NOTHING   
      end if
   end if
   SerInProgress=TRUE
   SerErr=NIL   
   InstanceList=NONE 
   x=des()
   if not equal(x,"ME") then  
      close_des()
      throw(Deserialize_Error)
      pop_stack()
      return NOTHING   
   end if
   x=des()      
   index=find(x,VERSION_LIST)
   if index=NIL then  
      close_des()
      throw(Deserialize_Error)
      pop_stack()
      return NOTHING
   end if 
   if index<=SV2 then
      SVersion=2
   else
      SVersion=1   
   end if
   x=des() 
   close_des()
   SerInProgress=FALSE
   if SerErr=NIL then 
      pop_stack()
      return x
   end if   
   for i=length(InstanceList) to 1 by -1 do
      thisInstance=InstanceList[i] 
      if instance_entity(thisInstance) then
         VOID=call_method(thisInstance,"delete",NONE)
         CurrentException[ProgramCounter]=NIL   
      end if        
   end for  
   throw(CLASS_HEAD&SerErr)               
   pop_stack()
   return NOTHING
end function  
  
-- Kernel classes:   

--Entity methods:
  
function Entity_new()
   integer targetClass
   targetClass=This[ProgramCounter][3]
   InstanceCounter+=1 
   if InstanceCounter>length(Instances) then
      Instances&=repeat(NIL,length(Instances))   
   end if 
   Instances[InstanceCounter]=PropertyDefault[targetClass][INSTANCE]
   return {MARKER,InstanceCounter,targetClass}
end function 
MethodName[ENTITY_ID][CLASS]&={"new"}
MethodSource[ENTITY_ID][CLASS]&=ENTITY_ID 
MethodAccess[ENTITY_ID][CLASS]&=PROTECTED
MethodParamCount[ENTITY_ID][CLASS]&=0                                 
MethodID[ENTITY_ID][CLASS]&=routine_id("Entity_new") 
MethodReturn[ENTITY_ID][CLASS]&=NIL

function Entity_clone()
   integer targetClass
   sequence target 
   target=This[ProgramCounter]
   targetClass=target[3]
   InstanceCounter+=1 
   if InstanceCounter>length(Instances) then
      Instances&=repeat(NIL,length(Instances))   
   end if 
   Instances[InstanceCounter]=Instances[target[2]]
   return {MARKER,InstanceCounter,targetClass}
end function                                  
MethodName[ENTITY_ID][INSTANCE]&={"clone"}
MethodSource[ENTITY_ID][INSTANCE]&=ENTITY_ID
MethodAccess[ENTITY_ID][INSTANCE]&=PROTECTED
MethodParamCount[ENTITY_ID][INSTANCE]&=0
MethodID[ENTITY_ID][INSTANCE]&=routine_id("Entity_clone") 
MethodReturn[ENTITY_ID][INSTANCE]&=NIL  

procedure Entity_delete()
   Instances[This[ProgramCounter][2]]=NIL
end procedure                                  
MethodName[ENTITY_ID][INSTANCE]&={"delete"}
MethodSource[ENTITY_ID][INSTANCE]&=ENTITY_ID 
MethodAccess[ENTITY_ID][INSTANCE]&=PROTECTED
MethodParamCount[ENTITY_ID][INSTANCE]&=0
MethodID[ENTITY_ID][INSTANCE]&=-routine_id("Entity_delete")-2 
MethodReturn[ENTITY_ID][INSTANCE]&={NOTHING}
-- End Entity methods


global constant Method_Wrapper=class("Method_Wrapper",Entity,NONE)

property("call_data",INSTANCE,PRIVATE,PRIVATE,NONE)

function Method_Wrapper_new(object target,object name) 
   integer access,caller,index,pc,source,targetClass,targetType 
   sequence new 
   call_proc(VALIDATE,{target,"entity",REQUIRED})
   call_proc(VALIDATE,{name,"identifier",REQUIRED})
   if CurrentException[ProgramCounter]>NIL then return NOTHING end if
   targetClass=target[3]  
   if target[2]!=CLASS_FLAG then
      targetType=INSTANCE
   else
      targetType=CLASS    
   end if
   index=find(name,MethodName[targetClass][targetType])
   if index=NIL then
      throw(Type_Check_Failure)
      return NOTHING      
   end if
   access=MethodAccess[targetClass][targetType][index] 
   pc=ProgramCounter-1 
   while TRUE do
      if access=PUBLIC then exit end if 
      caller=Caller[pc]      
      source=MethodSource[targetClass][targetType][index]
      if find(caller,SuperclassList[source]) then exit end if
      if access=PROTECTED then             
         if find(caller,SuperclassList[targetClass]) then exit end if
      end if
      if find(METHOD_WRAPPER_ID,SuperclassList[caller]) then
         pc-=1 
      else   
         throw(Type_Check_Failure)
         return NOTHING   
      end if
   end while 
   new=call_super(NONE) 
   set_property(new,"call_data",
   {target,name,MethodSource[targetClass][targetType][index],
   MethodParamCount[targetClass][targetType][index],
   MethodID[targetClass][targetType][index],
   MethodReturn[targetClass][targetType][index]})                
   return new
end function
method("new",CLASS,PUBLIC,2,routine_id("Method_Wrapper_new"))  

super_method("clone",INSTANCE)

super_method("delete",INSTANCE) 

procedure Method_Wrapper_deserialize() 
   integer targetClass,targetType,index
   sequence target
   object call_data
   call_data=get_property(this(),"call_data")
   if atom(call_data) then return end if
   while TRUE do
      target=call_data[1]
      if not entity(target) then exit end if 
      targetClass=target[3]
      if target[2]=CLASS_FLAG then
         targetType=CLASS
      else
         targetType=INSTANCE   
      end if
      index=find(call_data[2],MethodName[targetClass][targetType])
      if index=NIL then exit end if 
      call_data[3]=MethodSource[targetClass][targetType][index]
      call_data[4]=MethodParamCount[targetClass][targetType][index]
      call_data[5]=MethodID[targetClass][targetType][index]
      call_data[6]=MethodReturn[targetClass][targetType][index]
      set_property(this(),"call_data",call_data)
      return
   end while
   set_property(this(),"call_data",NIL)   
end procedure
method("deserialize",INSTANCE,PRIVATE,0,
{routine_id("Method_Wrapper_deserialize")})

function Method_Wrapper_call(sequence params) 
   integer id
   sequence me,target
   object call_data,ret             
   me=This[ProgramCounter]
   call_data=get_property(me,"call_data")
   if atom(call_data) then return NOTHING end if
   target=call_data[1]
   if deleted_entity(target) then 
      set_property(me,"call_data",NIL)
      return NOTHING
   end if 
   id=call_data[5]
   if id=BAD_METHOD then return call_data[6] end if
   params=wrap_param(params,call_data[4])  
   push_stack(call_data[3],target,call_data[2])
   if id<0 then
      ret=call_data[6]
      call_proc(-id-2,params)
   else
      ret=call_func(id,params)   
   end if
   pop_stack()
   return ret   
end function 
method("call",INSTANCE,PUBLIC,-1,routine_id("Method_Wrapper_call"))

end_class()    
METHOD_WRAPPER=Method_Wrapper  
METHOD_WRAPPER_ID=METHOD_WRAPPER[3]

constant Program=class("Program",Entity,NONE) 

null_method("main",CLASS,PRIVATE,NOTHING)
null_method("new",CLASS,PROTECTED,NOTHING)

end_class()

PROGRAM=Program  


end_package()  




-- Parameter validation: 
global procedure register_type(object name,object ID)
   if Context!=INITIALIZATION then
      fatal_error("register_type()\nNot allowed.")      
   end if
   if not identifier(name) then
      fatal_error("register_type()\nInvalid type name.")   
   end if 
   if length(CurrentPackage)>0 then
      name=CurrentPackage&"."&name   
   end if
   if find(name,TypeName) then
      fatal_error("register_type()\nType "&name
      &" duplicates existing type name.")   
   end if
   if not integer(ID) or ID<0 then
      fatal_error("register_type()\nType "&name
      &" has invalid routine id.")   
   end if 
   TypeName&={name}
   TypeID&=ID
end procedure

constant BOOLEAN_VALUES={TRUE,FALSE}

global type boolean(object x)
   return find(x,BOOLEAN_VALUES)!=NIL   
end type
register_type("boolean",routine_id("boolean"))

type integer_t(object x)
   return integer(x)
end type
register_type("integer",routine_id("integer_t"))

type atom_t(object x)
   return atom(x)
end type
register_type("atom",routine_id("atom_t"))

type object_t(object x)
   if entity(x) then return FALSE end if 
   if deleted_entity(x) then return FALSE end if
   if equal(x,NOTHING) then return FALSE end if
   return TRUE
end type      
register_type("object",routine_id("object_t"))

type sequence_t(object x)
   if atom(x) then return FALSE end if
   return object_t(x)
end type
register_type("sequence",routine_id("sequence_t"))  

register_type("identifier",routine_id("identifier"))

register_type("entity",routine_id("entity"))

register_type("instance",routine_id("instance_entity")) 

register_type("class",routine_id("class_entity"))  

register_type("string",routine_id("string")) 

type anything(object x)
   return object(x)
end type
register_type("anything",routine_id("anything"))  


function type_spec(object x) 
   integer index
   object x1,x2  
   if atom(x) then return NIL end if
   index=resolve_name(x,TRUE)
   if index!=NIL then return index end if
   if length(x)!=2 then return NIL end if  
   x1=x[1]  
   index=find(x1,ENTITY_TYPES) 
   if index=NIL then return NIL end if 
   index+=6 
   x2=valid_class(x[2],NIL)
   if atom(x2) then return NIL end if
   return -index
end function                                            

global type type_specifier(object x)
   return type_spec(x)!=NIL   
end type     
register_type("type_specifier",routine_id("type_specifier"))  
 
global procedure validate(object param,object typeSpec,object required)
   integer index,paramClass,typeClass
   if Context!=METHOD then
      fatal_error("validate()\nNot allowed.")
   end if
   if CurrentException[ProgramCounter]>NIL then return end if 
   if equal(param,NOTHING) then
      if not equal(required,OPTIONAL) then throw(Missing_Parameter) end if
      return   
   end if
   index=type_spec(typeSpec)
   if index=0 then 
      throw(Invalid_Type)   
   elsif index>0 then
      if equal(call_func(TypeID[index],{param}),FALSE) then 
         throw(Type_Check_Failure)
      end if                      
   else
      if equal(call_func(TypeID[-index],{param}),FALSE) then
         throw(Type_Check_Failure)   
      else
         typeClass=typeSpec[2][3] 
         paramClass=param[3]
         if find(typeClass,SuperclassList[paramClass]) then return end if
         if find(typeClass,InterfaceList[paramClass]) then return end if
         throw(Type_Check_Failure)
      end if
   end if
end procedure

VALIDATE=routine_id("validate")

function crash(object x) 
   integer fn  
   sequence mess
   VOID=x 
   if CrashFromFatal=FALSE then 
      FatalFromCrash=TRUE
      fn=open("ex.err","r")
      if fn=-1 then
         fatal_error("Euphoria error see above.") 
      else            
         VOID=gets(fn)
         mess=gets(fn)  
         while mess[1]=' ' do
            mess=mess[2..$]   
         end while
         while find(mess[$]," \r\n") do
            mess=mess[1..$-1]   
         end while
         mess[1]=mess[1]-(mess[1]>='a' and mess[1]<='z')*('a'-'A')
         if mess[$]!='.' then
            mess&='.'   
         end if
         close(fn)  
         clear_screen()
         fatal_error(mess) 
      end if
   end if
   FatalInProgress=TRUE
   terminate_program() 
   return NIL
end function

machine_proc(66,routine_id("crash"))  -- crash_routine

-- Task switching

procedure eu_task_yield()
	task_yield()
end procedure

sequence taskid taskid=NONE
sequence stack stack=NONE

global procedure task_yield()
        integer index,newindex
	atom self
	sequence entry
	self=task_self()
	index=find(self,taskid)
	if index=0 then
		taskid&=self
		stack&=0
		index=length(taskid)
	end if
	stack[index]={ProgramCounter,Context,Caller,This,ThisMethod}
	eu_task_yield()
	self=task_self()
	newindex=find(self,taskid)
	if newindex=index then return end if
	if newindex=0 then
		ProgramCounter=0
		Context=INITIALIZATION
		Caller=NONE
                This=NOTHING
		ThisMethod=NONE
		return 
	end if
	entry=stack[newindex]
	ProgramCounter=entry[1]
	Context=entry[2]
  	Caller=entry[3]
        This=entry[4]
        ThisMethod=entry[5]
end procedure
	        